#include <sea/kernel.h>
#include <sea/mm/vmm.h>
#include <sea/loader/elf-x86.h>
#include <sea/asm/system.h>
#include <sea/vsprintf.h>
/* This function's design is based off of JamesM's tutorials. 
 * Yes, I know its bad. But it works okay. */
void move_stack(void *start, unsigned int sz)
{
	unsigned i;
	for(i=(unsigned int)start+sz; i >= (unsigned int)start - sz*2;i -= 0x1000) {
		mm_vm_map(i, mm_alloc_physical_page(), PAGE_PRESENT | PAGE_WRITE | PAGE_USER, MAP_CRIT);
		memset((void *)i, 0, 0x1000);
	}
	unsigned pd_addr;
	u32int old_stack_pointer;
	u32int old_base_pointer;
	u32int offset = (u32int)start - initial_boot_stack;
	u32int new_base_pointer, new_stack_pointer, tmp, *tmp2;
	asm("mov %%cr3, %0" : "=r" (pd_addr));
	asm("mov %0, %%cr3" : : "r" (pd_addr)); 
	asm("mov %%esp, %0" : "=r" (old_stack_pointer));
	asm("mov %%ebp, %0" : "=r" (old_base_pointer));
	new_stack_pointer = old_stack_pointer + offset;
	new_base_pointer  = old_base_pointer  + offset; 
	memcpy((void*)new_stack_pointer, (void*)old_stack_pointer, initial_boot_stack-old_stack_pointer);
	
	for(i = (u32int)start; i > ((u32int)start-sz); i -= 4)
	{
		tmp = * (u32int*)i;
		if (( old_stack_pointer < tmp) && (tmp < initial_boot_stack))
		{
			tmp = tmp + offset;
			tmp2 = (u32int*)i;
			*tmp2 = tmp;
		}
	}
	
	asm("mov %0, %%esp" : : "r" (new_stack_pointer));
	asm("mov %0, %%ebp" : : "r" (new_base_pointer));
}

void setup_kernelstack()
{
	printk(1, "[stack]: Relocating stack\n");
	move_stack((void*)STACK_LOCATION, STACK_SIZE);
}

void arch_cpu_print_stack_trace(unsigned int MaxFrames)
{
	unsigned int * ebp = &MaxFrames - 2;
	kprintf("    ADDR      MODULE FUNCTION\n");
	for(unsigned int frame = 0; frame < MaxFrames; ++frame)
	{
		if((kernel_state_flags&KSF_MMU) && !mm_vm_get_map((addr_t)ebp, 0, 1)) break;
		unsigned int eip = ebp[1];
		if(eip == 0)
			break;
		ebp = (unsigned int *)(ebp[0]);
		const char *name = elf32_lookup_symbol(eip, &kernel_elf);
		char *modname = 0;
		if(!name)
			name = loader_lookup_module_symbol(eip, &modname);
		if(name) kprintf(" <%8.8x> %8s %s\n", eip, modname ? modname : "kernel", name);
	}
}

void arch_cpu_copy_fixup_stack(addr_t new, addr_t old, size_t length)
{
	memcpy((void *)new, (void *)old, length);
	int offset=0;
	offset = new-old;
	addr_t i;
	for(i = (u32int)new+(length-4); i >= (u32int)new; i -= 4)
	{
		u32int tmp = * (u32int*)i;
		if (tmp >= old && tmp < old+length)
		{
			tmp = tmp + offset;
			u32int *tmp2 = (u32int*)i;
			*tmp2 = tmp;
		}
	}
}

